/*
** reaper_talkback
** RME TotalMix Talkback control
** Copyright (C) AZ (www.azslow.com)
**    Based on code from Cockos Incorporated
** License: LGPL.
*/

#define _WINSOCK_DEPRECATED_NO_WARNINGS // inet_addr

#define WIN32_LEAN_AND_MEAN             // Exclude rarely-used stuff from Windows headers
// Windows Header Files:
#include <windows.h>
#include <WinSock2.h>

// TODO: reference additional headers your program requires here
#include "reaper_plugin.h"

#define REAPERAPI_MINIMAL
#define REAPERAPI_WANT_plugin_register
#define REAPERAPI_WANT_RefreshToolbar
#define REAPERAPI_IMPLEMENT
#include "reaper_plugin_functions.h"

// !!! Should match gAction
typedef enum {
	MUTE_ON_PLAYBACK  = 0,
	TOGGLE_MUTE
} ACTION_IDX;

static void RefreshToggle(ACTION_IDX idx);


class TalkBack
{
	SOCKET m_sendsock;
	struct sockaddr_in m_sendaddr;
	char m_msg[1024];
	char m_err[256]; // set in case we could not create socket

	bool m_mute_on_playback;   // TRUE when we want mute during playback

	// runtime
	bool m_talkback_muted; // TRUE when we think current TotalMix state is "Muted"
	bool m_playback;  // play && !pause
	bool m_recording; // recording && !pause

	CRITICAL_SECTION m_cs;

	bool _ToggleTalkBack()
	{
		if (m_sendsock == INVALID_SOCKET)
			return FALSE;
		m_talkback_muted = !m_talkback_muted; // be optimistic :)

		// "#bundle" (8) + time (8) + msg_len (4) + msg_body
		//const char *addr = "/1/mainTalkback"; // useless for several mics..
		const char *addr = "/3/muteGroups/4/1"; // mute group 1
		int addr_len = (int)strlen(addr) + 1;
		int aligned_addr_len = (addr_len + 3) & 0xfc;
		int msg_len = 20 + aligned_addr_len + 4 + 4;  // #bundle (8) + time (8) + size (4) + aligned addr + aligned ",f" + float (4)
		memset(m_msg, 0, sizeof(m_msg));
		memcpy(m_msg, "#bundle", 8);
		// time = 0
		*((int *)(m_msg + 16)) = htonl(aligned_addr_len + 4 + 4); // 4 bytes aligned addr + 4 bytes aligned ",f" + 4 bytes float
		memcpy(m_msg + 20, addr, addr_len);
		memcpy(m_msg + 20 + aligned_addr_len, ",f", 3);
		union {
			int i;
			float f;
		} arg;
		arg.f = 1.; // Toggle
		*((int *)(m_msg + 20 + aligned_addr_len + 4)) = htonl(arg.i);

		sendto(m_sendsock, m_msg, msg_len, 0, (struct sockaddr*)&m_sendaddr, sizeof(m_sendaddr));
		return TRUE;
	}


public:
	TalkBack() 
	{

		m_sendsock = socket(AF_INET, SOCK_DGRAM, 0);
		if (m_sendsock == INVALID_SOCKET) {
			FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM, NULL, WSAGetLastError(), 0, m_err, sizeof(m_err), NULL);
			//MessageBoxA(NULL, m_err, "TalkBack", MB_OK | MB_ICONERROR);
		}

		m_sendaddr.sin_family = AF_INET;
		m_sendaddr.sin_addr.s_addr = inet_addr("127.0.0.1");
		m_sendaddr.sin_port = htons(7001);

		m_mute_on_playback = FALSE;

		m_talkback_muted = FALSE; // we do not really know it without asking Totalmix, assume it is off.
		m_playback = FALSE;
		m_recording = FALSE;


		InitializeCriticalSection(&m_cs);
	}

	~TalkBack()
	{
		DeleteCriticalSection(&m_cs);
		if (m_sendsock != INVALID_SOCKET)
		{
			shutdown(m_sendsock, 2 /* SD_BOTH */);
			closesocket(m_sendsock);
			m_sendsock = INVALID_SOCKET;
		}
	}


	void SetPlayState(bool play, bool pause, bool rec)
	{
		play &= !pause;
		rec &= !pause;
		EnterCriticalSection(&m_cs);
		if (play != m_playback || rec != m_recording)
		{
			m_playback = play;
			m_recording = rec;
			bool should_mute = (play && m_mute_on_playback) || rec;
			if (m_talkback_muted != should_mute)
				_ToggleTalkBack();
		}
		LeaveCriticalSection(&m_cs);
	}

	void ToggleTalkBack() {
		bool refresh;
		EnterCriticalSection(&m_cs);
		refresh = _ToggleTalkBack();
		LeaveCriticalSection(&m_cs);
		if (refresh)
			RefreshToggle(TOGGLE_MUTE);
	}

	void ToggleMuteOnPlayback() {
		bool refresh = FALSE;
		EnterCriticalSection(&m_cs);
		m_mute_on_playback = !m_mute_on_playback;
		bool should_mute = (m_playback && m_mute_on_playback) || m_recording;
		if (m_talkback_muted != should_mute)
			refresh = _ToggleTalkBack();
		LeaveCriticalSection(&m_cs);
		if (refresh)
			RefreshToggle(TOGGLE_MUTE);
		RefreshToggle(MUTE_ON_PLAYBACK);
	}

	int isMuteOnPlayback() {
		return m_mute_on_playback ? 1 : 0; // no reason to lock
	}

	int isMuted() {
		return m_talkback_muted ? 1 : 0; // no reason to lock
	}

	const char *getErr() {
		return m_err;
	}
};

static TalkBack *gTalkBack = NULL;


static void mute_on_playback(int cmd, int flag) {
	if (gTalkBack)
		gTalkBack->ToggleMuteOnPlayback();
}

static int mute_on_playback_state(int cmd) {
	if (gTalkBack)
		return gTalkBack->isMuteOnPlayback();
	return -1;
}

static void toggle_mute(int cmd, int flag) {
	if (gTalkBack)
		gTalkBack->ToggleTalkBack();
}

static int toggle_mute_state(int cmd) {
	if (gTalkBack)
		return gTalkBack->isMuted();
	return -1;
}

static struct Action
{
	gaccel_register_t ac; // ac.accel.cmd should be the action id
	const char *id;
	void(*run_command)(int id, int flag);
	int(*toggle_state)(int id);
} gAction[] =
{
	{ { { 0, 0, 0 }, "TotalMix: Mute on playback"}, "TM_MUTEONPLAYBACK", mute_on_playback, mute_on_playback_state},
	{ { { 0, 0, 0 }, "TotalMix: Toggle mute" }, "TM_TOGGLEMUTE", toggle_mute, toggle_mute_state },
	{ { { 0, 0, 0 }, NULL }, NULL, NULL, NULL } // the end of list
};

static void RefreshToggle(ACTION_IDX idx) {
	int cmd = gAction[idx].ac.accel.cmd;
	if (cmd && RefreshToolbar)
		RefreshToolbar(cmd);
}

class CSurf_TalkBack : public IReaperControlSurface
{
	char configtmp[1024];

public:
	CSurf_TalkBack(const char *type_string, int *errStats) 
	{
		if (errStats && gTalkBack && gTalkBack->getErr()[0])
			*errStats = 2;
	}
	
	~CSurf_TalkBack() 
	{

	}

	const char *GetTypeString() { return "TALKBACK"; }

	const char *GetDescString() { return "RME TotalMix Talkback"; }

	const char *GetConfigString() // string of configuration data
	{
		strcpy_s(configtmp, "0");
		return configtmp;
	}

	void SetPlayState(bool play, bool pause, bool rec)
	{
		if (gTalkBack)
			gTalkBack->SetPlayState(play, pause, rec);
	}
};

static IReaperControlSurface *createFunc(const char *type_string, const char *configString, int *errStats)
{

	return new CSurf_TalkBack(configString, errStats);
}

static HWND configFunc(const char *type_string, HWND parent, const char *initConfigString)
{
	return NULL;
}


reaper_csurf_reg_t csurf_talkback_reg =
{
	"TALKBACK",
	"RME TotalMix Talkback",
	createFunc,
	configFunc,
};

static bool command_hook(int cmd, int flag) {
	for (struct Action *a = gAction; a->id; ++a) {
		if (a->ac.accel.cmd == cmd && a->run_command) {
			a->run_command(cmd, flag);
			return TRUE;
		}
	}
	return FALSE;
}

static int toggle_hook(int cmd) {
	for (struct Action *a = gAction; a->id; ++a) {
		if (a->ac.accel.cmd == cmd && a->toggle_state)
			return a->toggle_state(cmd);
	}
	return -1;
}

extern "C"
{

	REAPER_PLUGIN_DLL_EXPORT int REAPER_PLUGIN_ENTRYPOINT(REAPER_PLUGIN_HINSTANCE hInstance, reaper_plugin_info_t *rec)
	{
		if (!rec || rec->caller_version != REAPER_PLUGIN_VERSION || !rec->GetFunc) {
			if (gTalkBack) {
				delete gTalkBack;
				gTalkBack = NULL;
			}
			return 0;
		}

		if (REAPERAPI_LoadAPI(rec->GetFunc))
			return 0;

		if (!gTalkBack) {
			WSADATA wd;
			WSAStartup(MAKEWORD(2, 2), &wd);
			gTalkBack = new TalkBack();
		}

		rec->Register("csurf", &csurf_talkback_reg);
		rec->Register("hookcommand", command_hook);
		rec->Register("toggleaction", toggle_hook);
		for (struct Action *a = gAction; a->id; ++a) {
			a->ac.accel.cmd = rec->Register("command_id", (void *)(a->id));
			if (a->ac.accel.cmd)
				rec->Register("gaccel", (void *)&(a->ac));
		}

		return 1;

	}

};

BOOL APIENTRY DllMain(HMODULE hModule,
	DWORD  ul_reason_for_call,
	LPVOID lpReserved
)
{
	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
	}
	return TRUE;
}



